source('modules/Classes/Transition.R')
source('modules/Classes/PayoffFunction.R')

#Sources: 
#  "Avoiding the curse of dimensionality in dynamic stochastic games"
#  "Quality Ladder Model Notes" in Mendeley

DPEngine <- 
  setRefClass(
    'DPEngine',
    fields = list(
      eigen = 'numeric',
      bellman.count = 'numeric',
      theta = 'list',
      P = 'Transition',
      pi = 'PayoffFunction',
      x = 'data.frame',
      V = 'data.frame',
      W = 'data.frame'
    ),
    methods = list(
      initialize_DP = function(beta){
        theta <<- 
          list(
            beta = beta,
            M = 18,
            m = 5, 
            c = 5,
            alpha = 3, 
            delta = 0.7
          )
        
        pi <<-
          PayoffFunction$new(theta = theta) %T>% {
            .$calc_profit()
          }
        
        P <<-
          Transition$new(
            theta = theta
          )
        
        reset_system()
      },
      
      ####Basic Functions####
      update_x = function(){
        x <<-
          list(
            W %>% 
              fl(omega.i. == omega.i) %>% 
              sl(-omega.i.),
            W %>% 
              fl(omega.i. == omega.i - 1) %>% 
              rn(w.i.1 = w.i.) %>% 
              sl(-omega.i.),
            W %>% 
              fl(omega.i. == omega.i + 1) %>% 
              rn(w.i..1 = w.i.) %>% 
              sl(-omega.i.)
          ) %>% 
          rd(lj) %>% 
          mt(
            x.i = 
              ifelse(
                omega.i == 1,
                (sqrt(pmax(1, theta$beta * theta$alpha * (1 - theta$delta) * (w.i..1 - w.i.))) - 1)/theta$alpha,
                ifelse(
                  omega.i == theta$M,
                  (sqrt(pmax(1, theta$beta * theta$alpha * theta$delta * (w.i. - w.i.1))) - 1)/theta$alpha,
                  (sqrt(pmax(1, theta$beta * theta$alpha * (1 - theta$delta) * (w.i..1 - w.i.) + theta$beta * theta$alpha * theta$delta * (w.i. - w.i.1))) - 1)/theta$alpha
                )
              )
          ) %>% 
          sl(omega.i, omega.j, x.i)
      },
      
      update_P = function(){
        P$calc_tran(x)
      },
      
      update_V = function(porteus = FALSE){
        if(porteus){
          F.0 <- get_full_tran_mat()
          alpha <- theta$beta * min(diag(F.0))
          F.porteus <- (F.0 - (alpha/theta$beta) * diag(ncol(F.0)))/(1 - alpha)
          
          V <<-
            list(
              pi$profit,
              x,
              V %>% 
              mt(v.i. = theta$beta * F.porteus %*% v.i)
            ) %>% 
            rd(ij) %>% 
            mt(v.i = (profit - x.i)/(1-alpha) + v.i.) %>% 
            sl(omega.i, omega.j, v.i)
        } else {
          V <<-
            list(
              pi$profit,
              x,
              ij(W, P$tran) %>% 
                gb(omega.i, omega.j) %>% 
                sm(v.i. = theta$beta * sum(w.i. * prob.i))
            ) %>% 
            rd(ij) %>% 
            mt(v.i = profit - x.i + v.i.) %>% 
            sl(omega.i, omega.j, v.i)
        }
      },
      
      update_W = function() {
        W <<-
          P$tran %>% 
          flip_agents() %>% 
          ij(
            rn(
              V,
              omega.i. = omega.i, 
              omega.j. = omega.j
            )
          ) %>% 
          gb(omega.i, omega.j, omega.i.) %>% 
          sm(w.i. = sum(v.i * prob.j)) %>% 
          ug 
      },
      
      ####Solution operators####
      operator_pakes_iteration = function(porteus = FALSE){
        update_x()
        update_P()
        update_V(porteus)
        update_W()
      },
      
      ####DP solvers####
      solve_pakes = function(delta = 10^-10, relative.value = FALSE, porteus = FALSE){
        dif <- Inf
        while(dif > delta){
          bellman.count <<- bellman.count + 1
          print(dif)
          V.l <- V
          operator_pakes_iteration(porteus)
          if(relative.value) V$v.i <<- V$v.i - V$v.i[1]
          dif <- max(abs(V.l$v.i - V$v.i))
        }
      },
      
      ####Helpers####
      flip_agents = function(.){
        rename_all(., ~{
          sub('\\.i', '.I', .) %>% 
            sub('\\.j', '.i', .) %>% 
            sub('\\.I', '.j', .)
        })
      },
      
      get_full_tran_mat = function(){
        ij(
          P$tran,
          flip_agents(P$tran)
        ) %>% 
          mt(prob = prob.i * prob.j) %>% 
          sl(-c(prob.i, prob.j)) %>% 
          dc(
            omega.i + omega.j ~ omega.i. + omega.j.,
            value.var = 'prob'
          ) %>% 
          sl(-c(omega.i, omega.j)) %>% 
          as.matrix
      },
      
      get_eigen = function(){
        eigen <<- 
          get_full_tran_mat() %>% 
          eigen %>% {
            .$values
          } %>% 
          Mod
      },
      
      reset_system = function(){
        bellman.count <<- 0
        
        W <<-
          expand.grid(
            omega.i = seq(theta$M),
            omega.j = seq(theta$M),
            omega.i. = seq(theta$M)
          ) %>% 
          mt(w.i. = 1)
        
        V <<-
          expand.grid(
            omega.i = seq(theta$M),
            omega.j = seq(theta$M)
          ) %>% 
          mt(v.i = 1)
        
        x <<-
          rn(V, x.i = v.i)
          
        update_P()
      }
    )
  )